
from bergen.auths.types import HerreConfig
from oauthlib.oauth2.rfc6749.errors import InvalidClientError
from bergen.auths.base import AuthError, BaseAuthBackend
from oauthlib.oauth2.rfc6749.clients.backend_application import BackendApplicationClient
from requests_oauthlib.oauth2_session import OAuth2Session 
from bergen.enums import ClientType
import logging
import time

logger = logging.getLogger(__name__)


class ArnheimBackendOauth(BaseAuthBackend):
    failedTries = 5
    auto_retry = True


    def __init__(self, config: HerreConfig, **kwargs) -> None:
        super().__init__(config , **kwargs)  

    def fetchToken(self, loop=None) -> str:
        # Connecting
        logger.info(f"Connecting to Arnheim at {self.token_url}")
        
        auth_client = BackendApplicationClient(client_id=self.client_id, scope=self.scope)
        oauth_session = OAuth2Session(client=auth_client, scope=self.scope)


        def fetch_token(thetry=0):
            try:
                
                token = oauth_session.fetch_token(token_url=self.token_url, client_id=self.client_id,
                client_secret=self.client_secret, verify=True)


                if "access_token" not in token:
                    raise Exception("No access token Provided")

                return token["access_token"]
            except InvalidClientError as e:
                raise e
            except Exception as e:
                if thetry == self.failedTries or not self.auto_retry: raise AuthError(f"Cannot connect to Arnheim instance on {self.token_url}: {e}")
                logger.error(f"Couldn't connect to the Arnheim Instance at {self.token_url}. Retrying in 2 Seconds")
                return fetch_token(thetry=thetry + 1)
      
        return fetch_token()



from ....messages.types import CANCEL_ASSIGN, CANCEL_PROVIDE
from ....messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from typing import Optional


class MetaModel(MessageMetaModel):
    type: str = CANCEL_ASSIGN

class DataModel(MessageDataModel):
    reference: str


class CancelAssignMessage(MessageModel):
    data: DataModel
    meta: MetaModel


#Base Types
EXCEPTION = "exception"
ALLOWANCE = "allowance"

ASSIGNATION = "assignation"
PROVISION = "provision"
ASSIGNATION_REQUEST = "assignation_request"
PROVISION_REQUEST = "provision_request"


ACTIVATE_POD = "activate_pod"
DEACTIVATE_POD ="deactivate_pod"


# ASIGN PATH
ASSIGN = "assign"
CANCEL_ASSIGN = "cancel_assign"


# Arkitekt
BOUNCED_ASSIGN = "bounced_assign"
BOUNCED_CANCEL_ASSIGN = "bounced_cancel_assign"

# Pod 
ASSIGN_DONE = "assign_done"
ASSIGN_PROGRES ="assign_progress"
ASSIGN_CRITICAL = "assign_fail"
ASSIGN_YIELD =  "assign_yield"
ASSIGN_RETURN = "assign_return"
ASSIGN_RETRY = "assign_retry"

# Reserve Path
RESERVE = "reserve"
CANCEL_RESERVE = "cancel_reserve"

# Arkitekt
BOUNCED_RESERVE = "bounced_reserve"
BOUNCED_CANCEL_RESERVE = "bounced_cancel_reserve"

# Provider
RESERVE_DONE = "reserve_done"
RESERVE_PROGRESS = "reserve_progress"
RESERVE_FAIL = "reserve_fail"
RESERVE_CRITICAL = "reserve_critical"


UNRESERVE = "unreserve"

# Provide Path
PROVIDE = "provide"
CANCEL_PROVIDE = "cancel_provide"

# Arkitekt
BOUNCED_PROVIDE = "bounced_provide"
BOUNCED_CANCEL_PROVIDE = "bounced_cancel_provide"

# Provider
PROVIDE_DONE = "provide_done"
PROVIDE_PROGRESS = "provide_progress"
PROVIDE_FAIL = "provide_fail"
PROVIDE_CRITICAL = "provide_critical"



UNPROVIDE = "unprovide"





from asyncio.futures import Future
from bergen.entertainer.actor import HostHelper
from bergen.messages.postman.assign.assign import AssignMessage
from bergen.messages.base import MessageModel
from bergen.messages.utils import expandToMessage
from bergen.messages.host import ActivatePodMessage, DeActivatePodMessage
from bergen.messages.postman.assign import BouncedAssignMessage, BouncedCancelAssignMessage, AssignCriticalMessage, AssignReturnMessage, AssignYieldsMessage, AssignProgressMessage
from bergen.schema import Pod
import json
from bergen.entertainer.base import BaseHost
import logging
import asyncio
import websockets

import asyncio
try:
    from asyncio import create_task
except ImportError:
    #python 3.6 fix
    create_task = asyncio.ensure_future


logger = logging.getLogger()


class WebsocketHelper(HostHelper):

    async def pass_yield(self, message: AssignMessage, value):
        yield_message = AssignYieldsMessage(data={"yields": value}, meta={"extensions": message.meta.extensions, "reference": message.meta.reference})
        await self.host.send_to_arkitekt(yield_message)

    async def pass_progress(self, message: AssignMessage, value, percentage=None):
        message = f'{percentage if percentage else "--"} : {value}'
        progress_message = AssignProgressMessage(data={"message": value, "level": "INFO"}, meta={"extensions": message.meta.extensions, "reference": message.meta.reference})
        await self.host.send_to_arkitekt(progress_message)
        pass

    async def pass_result(self,message: AssignMessage, value):
        return_message = AssignReturnMessage(data={"returns": value}, meta={"extensions": message.meta.extensions, "reference": message.meta.reference})
        await self.host.send_to_arkitekt(return_message)

    async def pass_exception(self,message: AssignMessage, exception):
        error_message = AssignCriticalMessage(data={"message": str(exception)}, meta={"extensions": message.meta.extensions, "reference": message.meta.reference})
        await self.host.send_to_arkitekt(error_message)


class WebsocketHost(BaseHost):
    helperClass = WebsocketHelper
    ''' Is a mixin for Our Bergen '''

    def __init__(self, host= None, port= None, protocol=None, auto_reconnect=True, auth: dict= None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.websocket_host = host
        self.websocket_port = port
        self.websocket_protocol = protocol
        self.token = auth["token"]
        
        self.auto_reconnect= auto_reconnect
        self.allowed_retries = 2
        self.current_retries = 0

        self.all_pod_assignments = {}

    async def connect(self):
        self.incoming_queue = asyncio.Queue()
        self.outgoing_queue = asyncio.Queue()
        self.tasks = []

        self.startup_task = create_task(self.startup())

    async def disconnect(self) -> str:
        for reference, task in self.assignments.items():
            if not task.done():
                logger.info(f"Cancelling Assignment {task}")
                task.cancel()

        if self.connection: await self.connection.close()

        if self.pending:
            for task in self.pending:
                task.cancel()

        logger.info("Entertainer disconnected")

        
    async def startup(self):
        try:
            await self.connect_websocket()
        except Exception as e:

            logger.error(f"Entertainer Connection failed {e}")
            self.current_retries += 1
            if self.current_retries < self.allowed_retries and self.auto_reconnect:
                sleeping_time = (self.current_retries + 1)
                logger.info(f"Retrying in {sleeping_time} seconds")
                await asyncio.sleep(sleeping_time)
                await self.startup()
            else:
                logger.error("No reconnecting attempt envisioned. Shutting Down!")
                return

        self.consumer_task = create_task(
            self.consumer()
        )

        self.producer_task = create_task(
            self.producer()
        )

        self.worker_task = create_task(
            self.workers()
        )

        done, self.pending = await asyncio.wait(
            [self.consumer_task, self.worker_task, self.producer_task],
            return_when=asyncio.FIRST_EXCEPTION
        )

        logger.error(f"Lost connection inbetween everything :( {[ task.exception() for task in done]}")
        logger.error(f'Reconnecting')

        for task in self.pending:
            task.cancel()

        if self.connection: await self.connection.close()
        self.current_retries = 0 # reset retries after one successfull connection
        await self.startup() # Attempt to ronnect again
        

    async def connect_websocket(self):

        uri = f"{self.websocket_protocol}://{self.websocket_host}:{self.websocket_port}/host/?token={self.token}"
        
        self.connection = await websockets.client.connect(uri)
        logger.info("Connecting as Entertainer")


    async def consumer(self):
        logger.warning(" [x] Awaiting Entertaining Calls")
        async for message in self.connection:
            logger.info(f"Incoming {message}")
            await self.incoming_queue.put(message)


    async def producer(self):
        while True:
            message = await self.outgoing_queue.get()
            await self.connection.send(message.to_channels())

            self.outgoing_queue.task_done()


    async def send_to_arkitekt(self, message: MessageModel):
        await self.outgoing_queue.put(message)
       

    def raise_exception(self, future: Future):
        if future.exception():
            raise future.exception()

    async def workers(self):
        while True:
            message_str = await self.incoming_queue.get()
            message = expandToMessage(json.loads(message_str))
            logger.info("Received Assignation")

            if isinstance(message, BouncedAssignMessage):
                assert message.data.pod is not None, "Received assignation that had no Pod???"
                assert message.data.pod in self.pod_actor_run_map, f"Pod not entertained {message.data.pod} {self.pod_actor_run_map}"
                await self.pod_actor_queue_map[message.data.pod].put(message)
                self.all_pod_assignments[message.meta.reference] = message.data.pod # Run in parallel

            if isinstance(message, BouncedCancelAssignMessage):
                if message.data.reference in self.all_pod_assignments: 
                    logger.info("Cancellation for task received. Canceling!")
                    task = self.assignments[message.data.reference]
                    if not task.done():
                        task.cancel()
                        logger.warn("Canceled Task!!")
                    else:
                        logger.warn("Task had already finished")
                else:
                    logger.error("Received Cancellation for task that was not in our tasks..")

            
            self.incoming_queue.task_done()


    async def deactivatePod(self, id, reference = "spmposmpe"):
        deactivate = DeActivatePodMessage(data={"pod": id}, meta={"reference": reference})
        await self.send_to_arkitekt(deactivate)
        logger.warn(f"Setting Pod inactive {reference}")

    async def activatePod(self, pod: Pod, reference):
        activate = ActivatePodMessage(data={"pod": pod.id}, meta={"reference": reference})
        await self.send_to_arkitekt(activate)


    
    




import asyncio
from oauthlib.oauth2.rfc6749.clients.base import Client

from pydantic.main import BaseModel
from bergen.auths.types import User
from typing import Dict
from bergen.wards.graphql.aiohttp import AIOHttpGraphQLWard
from bergen.enums import ClientType, HostProtocol, PostmanProtocol, ProviderProtocol
from bergen.logging import setLogging
from bergen.auths.base import BaseAuthBackend
from bergen.wards.base import BaseWard
from bergen.postmans.base import BasePostman
import logging
from threading import Thread


logger = logging.getLogger(__name__)
import os

def start_background_loop(loop: asyncio.AbstractEventLoop) -> None:
    asyncio.set_event_loop(loop)


class ArkitektConfig(BaseModel):
    secure: bool
    host: str
    port: int

    def __str__(self) -> str:
        return f"{'Secure' if self.secure else 'Insecure'} Connection to Arkitekt on {self.host}:{self.port}"



class BaseBergen:


    def __init__(self, auth: BaseAuthBackend, config: ArkitektConfig , 
            auto_negotiate=True,
            bind=True,
            log=logging.INFO,
            jupyter=False,
            name=None,
            force_sync = False,
            loop = None,
            client_type: ClientType = ClientType.CLIENT,
            **kwargs) -> None:
        
        if jupyter:
            setLogging(logging.ERROR)
        else:
            print(r"     _               _          _              ____ _ _            _    ")   
            print(r"    / \   _ __ _ __ | |__   ___(_)_ __ ___    / ___| (_) ___ _ __ | |_  ")
            print(r"   / _ \ | '__| '_ \| '_ \ / _ \ | '_ ` _ \  | |   | | |/ _ \ '_ \| __| ")
            print(r"  / ___ \| |  | | | | | | |  __/ | | | | | | | |___| | |  __/ | | | |_  ")
            print(r" /_/   \_\_|  |_| |_|_| |_|\___|_|_| |_| |_|  \____|_|_|\___|_| |_|\__| ")
            print(r"")
            setLogging(log)


        self.running_in_sync = force_sync
        
        self.loop = loop or asyncio.get_event_loop()
        if self.loop.is_running():
            if self.running_in_sync:
                logger.warn("force_insync within event-loop. unexpected errors might be happening")
                import nest_asyncio
                nest_asyncio.apply(self.loop)
            self.loop_is_running = True
        else:
            self.loop_is_running = False

        
        if bind: 
            # We only import this here for typehints
            from bergen.registries.arnheim import set_current_arnheim
            set_current_arnheim(self)

        self.auth = auth
        self.config = config
        self.name = name
        self.client_type = client_type


        self.token = self.auth.getToken()

        logger.info("We are authorized")
        logger.info(str(config))


        self.host = config.host
        self.port = config.port
        self.protocol = "https" if config.secure else "http"

        self._transcript = None
        self.identifierDataPointMap = {}
        self.identifierWardMap: Dict[str, BaseWard] = {}

        self._provider = None
        self._entertainer = None


        if auto_negotiate == True:
            if self.loop.is_running() and not self.running_in_sync: 
                pass
            else:
               self.negotiate()


        super().__init__() 

    def getLoopAndContext(self):
        return self.loop, self.running_in_sync    


    @property
    def transcript(self):
        assert self._transcript is not None, "We have to negotiate first with our"
        return self._transcript

    def getExtensionSettings(self, extension):
        assert extension in self.transcript.extensions, f"Arnheim seems to have no idea about this Extension {extension}"
        return self.transcript.extensions[extension]

    def getWardForIdentifier(self, identifier):
        if identifier in ["node","template","pod"]:
            return self.main_ward

        if self._transcript is None:
            if self.running_in_sync:
                raise Exception("Not negotiated Error: Please negotiate first or set auto_negotiate=True")
            else:
                raise Exception("You are running in event Loop: Please await self.negotiate_async first or run with 'async with Bergen(....)'")

        if identifier in self.identifierWardMap:
            return self.identifierWardMap[identifier]
        else:
            raise Exception(f"Couldn't find a Ward/Datapoint for Model {identifier}, this mostly results from importing a schema that isn't part of your arkitekts configuration ..Check Documentaiton")


    def getPostmanFromSettings(self, transcript):
        settings = transcript.postman

        if settings.type == PostmanProtocol.RABBITMQ:
            try:
                from bergen.postmans.pika import PikaPostman
                postman = PikaPostman(**settings.kwargs, loop=self.loop, client=self)
            except ImportError as e:
                logger.error("You cannot use the Pika Postman without installing aio_pika")
                raise e

        elif settings.type == PostmanProtocol.WEBSOCKET:
            try:
                from bergen.postmans.websocket import WebsocketPostman
                postman = WebsocketPostman(**settings.kwargs, loop=self.loop, client=self)
            except ImportError as e:
                logger.error("You cannot use the Websocket Postman without installing websockets")
                raise e

        else:
            raise Exception(f"Postman couldn't be configured. No Postman for type {settings.type}")

        return postman

    def getProviderFromSettings(self, transcript):
        settings = transcript.provider

        if settings.type == ProviderProtocol.WEBSOCKET:
            try:
                from bergen.provider.websocket import WebsocketProvider
                provider = WebsocketProvider(**settings.kwargs, loop=self.loop, client=self, name=self.name)
            except ImportError as e:
                logger.error("You cannot use the Websocket Provider without installing websockets")
                raise e

        else:
            raise Exception(f"Provider couldn't be configured. No Provider for type {settings.type}")

        return provider

    def getEntertainerFromSettings(self, transcript):
        settings = transcript.host

        if settings.type == HostProtocol.WEBSOCKET:
            try:
                from bergen.entertainer.websocket import WebsocketHost
                provider = WebsocketHost(**settings.kwargs, loop=self.loop, client=self)
            except ImportError as e:
                logger.error("You cannot use the Websocket Entertainer without installing websockets")
                raise e

        else:
            raise Exception(f"Entertainer couldn't be configured. No Entertainer for type {settings.type}")

        return provider
    
    async def negotiate_async(self):
        from bergen.constants import NEGOTIATION_GQL
        from bergen.registries.datapoint import get_datapoint_registry

        # Instantiate our Main Ward, this is only for Nodes and Pods
        self.main_ward = AIOHttpGraphQLWard(host=self.host, port=self.port, protocol=self.protocol, token=self.token, loop=self.loop)
        await self.main_ward.configure()

        # We resort escalating to the different client Type protocols
        logger.info(f"Negotiating to be a {self.client_type}")
        self._transcript = await NEGOTIATION_GQL.run_async(ward=self.main_ward, variables={"clientType": self.client_type, "name": self.name})
        

        #Lets create our different Wards 
        
        assert self._transcript.models is not None, "We apparently didnt't get any points"
        
        datapoint_registry = get_datapoint_registry()


        self.identifierDataPointMap = {model.identifier.lower(): model.point for model in self._transcript.models}
        self.identifierWardMap = {model.identifier.lower(): datapoint_registry.createWardForDatapoint(model.point, self) for model in self._transcript.models}

        logger.info("Succesfully registered Datapoints") 
        await datapoint_registry.configureWards()
        logger.info("Succesfully connected to Datapoints") 


        self.postman = self.getPostmanFromSettings(self._transcript)
        await self.postman.connect()

        if self.client_type in [ClientType.PROVIDER, ClientType.HOST]:
            logger.warn("We are connected as a Host")
            self._entertainer = self.getEntertainerFromSettings(self._transcript)
            await self._entertainer.connect()

        if self.client_type == ClientType.PROVIDER:
            logger.warn("We are connected as a Provider")
            self._provider = self.getProviderFromSettings(self._transcript)
            await self._provider.connect()



    async def disconnect_async(self, client_type=None):
        await self.main_ward.disconnect()

        if self.postman: await self.postman.disconnect()
        if self._provider: await self._provider.disconnect()
        if self._entertainer: await self._entertainer.disconnect()

        wards = self.identifierWardMap.values()
        for ward in wards:
            await ward.disconnect()


    def negotiate(self, client_type = None):
        self.loop.run_until_complete(self.negotiate_async())


    def getUser(self) -> User:
        return self.auth.getUser()


    def getExtensions(self, service):
        assert service in self._transcript.extensions, "This Service doesnt register Extensions on Negotiate"
        assert self._transcript.extensions[service] is not None, "There are no extensions registered for this Service and this App (see negotiate)"
        return self._transcript.extensions[service]
    
        
    def getWard(self) -> BaseWard:
        return self.main_ward

    def getPostman(self) -> BasePostman:
        return self.postman

    def _repr_html_(self):
        if not self._transcript: return """Unconnected Client"""
        return f"""
            <p> Arnheim Client <p>
            <table>
                <tr>
                    <td> Connected to </td> <td> {self.main_ward.name} </td>
                </tr>
            </table>

        """

    async def __aenter__(self):
        await self.negotiate_async()
        return self


    async def __aexit__(self,*args, **kwargs):
        await self.disconnect_async()

    
    @property
    def provider(self):
        if self._provider:
            return self._provider
        else:
            raise Exception("We are not in Provider mode")

    @property
    def entertainer(self):
        if self._entertainer:
            return self._entertainer
        else:
            raise Exception("We are not in Enterainer mode")


    def enable(self, *args, **kwargs):
        if self._provider:
            return self.provider.enable(*args, **kwargs)
        else:
            raise Exception("We are not in Provider Mode")

    def provide(self):
        return self.provider.provide()




from pydantic.main import BaseModel
from ....messages.types import  UNPROVIDE
from ....messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from typing import List, Optional



class UnProvideMessageMetaExtensionsModel(MessageMetaExtensionsModel):
    # Set by postman consumer
    progress: Optional[str]
    callback: Optional[str]

class UnProvideMetaModel(MessageMetaModel):
    type: str = UNPROVIDE
    extensions: Optional[UnProvideMessageMetaExtensionsModel]

class UnProvideDataModel(MessageDataModel):

    node: Optional[str] #TODO: Maybe not optional
    template: Optional[str]

class UnProvideMessage(MessageModel):
    data: UnProvideDataModel
    meta: UnProvideMetaModel


from ....messages.types import CANCEL_PROVIDE
from ....messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from typing import Optional


class CancelProvideMetaModel(MessageMetaModel):
    type: str = CANCEL_PROVIDE

class CancelProvideDataModel(MessageDataModel):
    reference: str


class CancelProvideMessage(MessageModel):
    data: CancelProvideDataModel
    meta: CancelProvideMetaModel




from abc import ABC, abstractmethod
from asyncio.events import AbstractEventLoop
from asyncio.futures import Future
from re import M
from bergen.messages.postman.assign.bounced_cancel_assign import BouncedCancelAssignMessage
from bergen.messages.postman.assign.bounced_assign import BouncedAssignMessage
import contextvars
from typing import Type
from bergen.types.node.ports import kwarg
from bergen.messages.postman.assign.assign import AssignMessage
import uuid

from pydantic.main import BaseModel
from bergen.utils import expandInputs, shrinkInputs, shrinkOutputs
from bergen.schema import Template
from bergen.constants import ACCEPT_GQL, OFFER_GQL, SERVE_GQL
import logging
import asyncio
from bergen.models import Node, Pod
import inspect
from aiostream import stream
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from functools import partial


logger = logging.getLogger()

class HostHelper(ABC):

    def __init__(self, host) -> None:
        self.host = host
        pass

    @abstractmethod
    async def pass_yield(self, message, value):
        pass

    @abstractmethod
    async def pass_progress(self, message, value, percentage=None):
        pass

    @abstractmethod
    async def pass_result(self, message, value):
        pass

    @abstractmethod
    async def pass_exception(self, message, exception):
        pass


assign_var = contextvars.ContextVar('assign', default=None)

class Actor:

    def __init__(self, pod: Pod, helper: HostHelper, loop=None) -> None:
        self.queue = asyncio.Queue()
        self.pod = pod
        self.helper = helper
        self.loop = loop
        self.assignments = {}
        pass

    async def on_provide(self):
        pass

    def check_if_assignation_cancelled(self, assign, future):
        print(future)

    async def run(self):
        ''' An infinitie loop assigning to itself'''
        try:
            logger.info(f"Run for {self.__class__.__name__} on Pod {self.pod.id}: Pod was Provided. Reserving")
            await self.on_provide()
            logger.info(f"Run for {self.__class__.__name__} on Pod {self.pod.id}: Resources were reservered. Waiting for Tasks until canceled")
            
            while True:
                message = await self.queue.get()
                logger.info(f"Assignation for {self.__class__.__name__} on Pod {self.pod.id}: {message}")

                if isinstance(message, BouncedAssignMessage):
                    logger.info(f"Assignation for {self.__class__.__name__} on Pod {self.pod.id}: {message}")
                    task = asyncio.create_task(self.on_assign(message))
                    task.add_done_callback(partial(self.check_if_assignation_cancelled, message))
                    self.assignments[message.meta.reference] = task

                elif isinstance(message, BouncedCancelAssignMessage):
                    logger.info(f"Assignation Cancellation for {self.__class__.__name__} on Pod {self.pod.id}: {message}")
                    if message.data.reference in self.assignments: 
                        logger.info("Cancellation for task received. Canceling!")
                        task = self.assignments[message.data.reference]
                        if not task.done():
                            task.cancel()
                            logger.warn("Canceled Task!!")
                        else:
                            logger.warn("Task had already finished")
                    else:
                        raise Exception("Assignment never was at this pod. something went wrong")

                else:
                    raise Exception(f"Type not known {message}")

                self.queue.task_done()


        except asyncio.CancelledError:
            logger.info(f"Run for {self.__class__.__name__} on Pod {self.pod.id}: Received Entertainment Cancellation. Unreserving")
            await self.on_unprovide()
            logger.info(f"Run for {self.__class__.__name__} on Pod {self.pod.id}: Unprovided. Ciao!")
            raise

        except Exception as e:
            logger.error(e)
            raise e

    async def on_unprovide(self):
        pass

    async def progress(self, value, percentage):
        message = assign_var.get()
        await self.helper.pass_progress(message, value, percentage)

    async def on_assign(self, assign: AssignMessage):
        args, kwargs = await expandInputs(node=self.pod.template.node, args=assign.data.args, kwargs=assign.data.kwargs)
        logger.info(args, kwargs)
        assign_var.set(assign)
        await self._assign(**{**args, **kwargs})
    
    async def __aenter__(self):
        await self.reserve()
        return self

    async def __aexit__(self):
        await self.unreserve()



class AsyncActor(Actor):

    def __init__(self, pod: Pod, helper: HostHelper) -> None:
        super().__init__(pod, helper)



class AsyncFuncActor(AsyncActor):

    async def assign(self, *args, **kwargs):
        raise NotImplementedError("Please provide a func or overwrite the assign method!")

    async def _assign(self, *args, **kwargs):
        message = assign_var.get()
        result = await self.assign(*args, **kwargs)
        try:
            shrinked_returns = await shrinkOutputs(self.pod.template.node, result)
            await self.helper.pass_result(message, shrinked_returns)
        except Exception as e:
            logger.error(e)
            await self.helper.pass_exception(message, e)




class AsyncGenActor(AsyncActor):

    async def assign(self):
        raise NotImplementedError("This needs to be overwritten in order to work")

    async def _assign(self, *args, **kwargs):
        message = assign_var.get()
        yieldstream = stream.iterate(self.assign(*args, **kwargs))
        async with yieldstream.stream() as streamer:
            async for result in streamer:
                lastresult = await shrinkOutputs(self.pod.template.node, result)
                await self.helper.pass_yield(message, lastresult)

        await self.helper.pass_result(message, lastresult)


class ThreadedFuncActor(Actor):
    nworkers = 5

    def __init__(self, pod: Pod, helper: HostHelper, loop) -> None:
        super().__init__(pod, helper, loop=loop)
        self.threadpool = ThreadPoolExecutor(self.nworkers)

    def progress(self, value, percentage):
        message = assign_var.get()
        logger.info(f'{percentage if percentage else "--"} : {value}')
        if self.loop.is_running():
            future = asyncio.run_coroutine_threadsafe(self.helper.pass_progress(message, value, percentage=percentage), self.loop)
            return future.result()
        else:
            self.loop.run_until_complete(self.helper.pass_progress(message, value, percentage=percentage))

    def assign(self):
        raise NotImplementedError("")   

    async def _assign(self, *args, **kwargs):
        message = assign_var.get()
        result = await self.loop.run_in_executor(self.threadpool, self.assign)

        returns = await shrinkOutputs(self.pod.template.node, result)
        await self.helper.pass_result(message, returns)





from bergen.auths.types import HerreConfig
from oauthlib.oauth2.rfc6749.clients.legacy_application import LegacyApplicationClient
from oauthlib.oauth2.rfc6749.clients.mobile_application import MobileApplicationClient
import requests
from requests_oauthlib.oauth2_session import OAuth2Session
from bergen.auths.base import BaseAuthBackend
from bergen.enums import ClientType

class ImplicitError(Exception):
    pass


class LegacyApplication(BaseAuthBackend):


    def __init__(self, config: HerreConfig, username=None, password=None, **kwargs) -> None:
        super().__init__(config, **kwargs)
        self.username = username
        self.password = password

    
    def fetchToken(self, loop=None) -> str:
        # Getting token

        self.legacy_app_client =  LegacyApplicationClient(self.client_id)
        if not self.username: self.username = input("Enter your username:    ")
        if not self.password: self.password = input("Password?               ")

        data = { "username": self.username, "password": self.password, "grant_type": "password", "scope": self.scope, "client_id": self.client_id, "client_secret": self.client_secret}

        url = self.token_url + "/"
        try:
            response = requests.post(url, data=data).json()
        except Exception as e:
            raise e


        if "access_token" in response:
            return response["access_token"]
        else:
            raise Exception(f"Wasn't authorized! {response}")



from .assign import AssignMessage
from .assign_critical import AssignCriticalMessage
from .assign_progress import AssignProgressMessage
from .assign_return import AssignReturnMessage
from .bounced_assign import BouncedAssignMessage
from .bounced_cancel_assign import BouncedCancelAssignMessage
from .cancel_assign import CancelAssignMessage
from .assign_yield import AssignYieldsMessage


from ....messages.generics import Token
from ....messages.types import BOUNCED_CANCEL_RESERVE
from ....messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from typing import Optional


class MetaExtensionsModel(MessageMetaExtensionsModel):
    # Set by postman consumer
    progress: Optional[str]
    callback: Optional[str]

class MetaModel(MessageMetaModel):
    type: str = BOUNCED_CANCEL_RESERVE
    extensions: MetaExtensionsModel
    token: Token

class DataModel(MessageDataModel):
    reference: str


class BouncedCancelReserveMessage(MessageModel):
    data: DataModel
    meta: MetaModel


from ....messages.generics import Token
from pydantic.main import BaseModel
from ....messages.types import  BOUNCED_ASSIGN
from ....messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from typing import List, Optional



class AssignParams(BaseModel):
    providers: Optional[List[str]]

class MetaExtensionsModel(MessageMetaExtensionsModel):
    # Set by postman consumer
    progress: Optional[str]
    callback: Optional[str]

class MetaModel(MessageMetaModel):
    type: str = BOUNCED_ASSIGN
    extensions: Optional[MetaExtensionsModel]
    token: Token

class DataModel(MessageDataModel):
    node: Optional[str] #TODO: Maybe not optional
    template: Optional[str]
    pod: Optional[str]

    reservation: Optional[str]

    params: Optional[AssignParams]

    args: Optional[dict]
    kwargs: Optional[dict]


class BouncedAssignMessage(MessageModel):
    data: DataModel
    meta: MetaModel


from enum import Enum
from pydantic.main import BaseModel
from ....messages.types import  ASSIGN_PROGRES, PROVIDE, PROVIDE_DONE
from ....messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from typing import List, Optional


class MetaExtensionsModel(MessageMetaExtensionsModel):
    # Set by postman consumer
    progress: Optional[str]
    callback: Optional[str]

class MetaModel(MessageMetaModel):
    type: str = ASSIGN_PROGRES
    extensions: Optional[MetaExtensionsModel]


class ProgressLevel(str, Enum):
    INFO = "INFO"
    DEBUG = "DEBUG"


class DataModel(MessageDataModel):
    level: ProgressLevel
    message: str

class AssignProgressMessage(MessageModel):
    data: DataModel
    meta: MetaModel




from abc import ABC, abstractmethod
from asyncio.events import AbstractEventLoop
from bergen.entertainer.actor import Actor, AsyncFuncActor, AsyncGenActor, ThreadedFuncActor
from bergen.messages.postman.assign.assign import AssignMessage
from bergen.provider.utils import createNodeFromActor, createNodeFromFunction

from pydantic.main import BaseModel
from bergen.types.node import ports
from typing import Dict, Tuple, TypedDict, Union
from bergen.utils import ExpansionError, expandInputs, shrinkOutputs
from bergen.schema import Template, NodeType
from bergen.constants import ACCEPT_GQL, OFFER_GQL, SERVE_GQL
import logging
import namegenerator
import asyncio
import websockets
from bergen.models import Node, Pod
import inspect
import sys
from aiostream import stream
from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor
from functools import partial
logger = logging.getLogger()
import copy
import json
from bergen.types.model import ArnheimModel
import uuid



class BaseHelper(ABC):

    def __init__(self, peasent) -> None:
        self.peasent = peasent
        pass

    @abstractmethod
    async def pass_yield(self, message, value):
        pass

    @abstractmethod
    async def pass_progress(self, message, value, percentage=None):
        pass

    @abstractmethod
    async def pass_result(self, message, value):
        pass

    @abstractmethod
    async def pass_exception(self, message, exception):
        pass


class AssignationHelper():

    def __init__(self, peasent_helper: BaseHelper, message: AssignMessage, loop: AbstractEventLoop = None) -> None:
        self.peasent_helper = peasent_helper
        self.message = message
        self.loop = loop
        pass

    @abstractmethod
    def progress(self, value, percentage=None):
        pass



class ThreadedAssignationHelper(AssignationHelper):

    def progress(self, value, percentage=None):
        logger.info(f'{percentage if percentage else "--"} : {value}')
        if self.loop.is_running():
            future = asyncio.run_coroutine_threadsafe(self.peasent_helper.pass_progress(self.message, value, percentage=percentage), self.loop)
            return future.result()
        else:
            self.loop.run_until_complete(self.peasent_helper.pass_progress(self.message, value, percentage=percentage))


class AsyncAssignationHelper(AssignationHelper):

    async def progress(self, value, percentage=None):
        logger.info(f'{percentage if percentage else "--"} : {value}')
        await self.peasent_helper.pass_progress(self.message, value, percentage=percentage)



async def shrink(node_template_pod, outputs):
    kwargs = {}
    if isinstance(node_template_pod, Node):
        kwargs = await shrinkOutputs(node=node_template_pod, outputs=outputs)

    if isinstance(node_template_pod, Template):
        kwargs = await shrinkOutputs(node=node_template_pod.node, outputs=outputs)

    if isinstance(node_template_pod, Pod):
        kwargs = await shrinkOutputs(node=node_template_pod.template.node, outputs=outputs)


    return kwargs



async def expand(node_template_pod, inputs):
    kwargs = {}
    if isinstance(node_template_pod, Node):
        kwargs = await expandInputs(node=node_template_pod, inputs=inputs)

    if isinstance(node_template_pod, Template):
        kwargs = await expandInputs(node=node_template_pod.node, inputs=inputs)

    if isinstance(node_template_pod, Pod):
        kwargs = await expandInputs(node=node_template_pod.template.node, inputs=inputs) 
    return kwargs



threadhelper = None
try: 
    threadhelper = asyncio.to_thread
except:
    logger.warn("Threading does not work below Python 3.9")



class PodPolicy(BaseModel):
    type: str



class OneExlusivePodPolicy(PodPolicy):
    type: str = "one-exclusive"

class MultiplePodPolicy(PodPolicy):
    type: str = "multiple"



def isactor(type):
    try:
        if issubclass(type, Actor):
            return True
        else:
            return False
    except Exception as e:
        return False



class BaseProvider:
    ''' Is a mixin for Our Bergen '''
    helperClass = None


    def __init__(self, *args, name = None, provider: int = None, loop=None, client=None) -> None:
        assert provider is not None, "Provider was set to none, this is weird!!"
        self.arkitekt_provider = provider
        self.name = name
        self.client = client
        self.loop = loop or asyncio.get_event_loop()

        self.template_actorClass_map = {}

        self.provisions = {}
        self.auto_provided_templates = [] # Templates that this provide will try to provide on startup


    def template(self, node: Node, policy: PodPolicy = MultiplePodPolicy(), auto_provide=False, on_provide=None,  on_unprovide=None, **implementation_details):

        def real_decorator(function_or_actor):
            assert callable(function_or_actor) or issubclass(function_or_actor, Actor), "Please only decorate functions or subclasses of Actor"
            # TODO: Check if function has same parameters as node

            template = OFFER_GQL.run(
            ward=self.client.main_ward,
            variables= {
                "node": node.id,
                "params": implementation_details,
                "policy": policy.dict()
            })


            if isactor(function_or_actor):
                actorClass = function_or_actor
            else:
                is_coroutine = inspect.iscoroutinefunction(function_or_actor)
                is_asyncgen = inspect.isasyncgenfunction(function_or_actor)
                is_function = inspect.isfunction(function_or_actor)

                if is_coroutine:
                    actorClass =  type(f"GeneratedActorNode{node.id}",(AsyncFuncActor,), {"assign": function_or_actor})
                elif is_asyncgen:
                    actorClass =  type(f"GeneratedActorNode{node.id}",(AsyncGenActor,), {"assign": function_or_actor})
                elif is_function:
                    actorClass = type(f"GeneratedActorNode{node.id}",(ThreadedFuncActor,), {"assign": function_or_actor})
                else:
                    raise Exception(f"Unknown type of function {function_or_actor}")
            
            self.template_actorClass_map[str(template.id)] = actorClass
            return actorClass

        return real_decorator


    def enable(self, allow_empty_doc=False, widgets={}, **implementation_details):
        """Enables the decorating function as a node on the Arnheim, you will find it as
        @provider/

        Args:
            allow_empty_doc (bool, optional): Allow the enabled function to not have a documentation. Will automatically downlevel the Node Defaults to False.
            widgets (dict, optional): Enable special widgets for the parameters. Defaults to {}.
        """


        assert self.name is not None, "We have no name provided, cannot put created Nodes in a Package without a"
        def real_decorator(function_or_actor):
            assert callable(function_or_actor) or issubclass(function_or_actor, Actor), "Please only decorate functions or subclasses of Actor"

            if isactor(function_or_actor):
                logger.info("Already is Actor. Creating Node")
                node = createNodeFromActor(function_or_actor, self.name, allow_empty_doc=allow_empty_doc, widgets=widgets)
            else:
                logger.info("Is Function. Creating Node and Wrapping")
                node = createNodeFromFunction(function_or_actor, self.name, allow_empty_doc=allow_empty_doc, widgets=widgets)
                
            # We pass this down to our truly template wrapper that takes the node and transforms it
            template_wrapper = self.template(node, **implementation_details)
            function = template_wrapper(function_or_actor)
            return function

        return real_decorator


    @abstractmethod
    async def connect(self) -> str:
        raise NotImplementedError("Please overwrite")

    @abstractmethod
    async def disconnect(self) -> str:
        raise NotImplementedError("Please overwrite")

    
    async def provideTemplate(self, reference: str, template_id: str):
        assert template_id in self.template_actorClass_map, f"There is no function registered for this template {template_id} not it {self.template_actorClass_map.keys()}"
            
        actor = self.template_actorClass_map[template_id]
        pod = await ACCEPT_GQL.run_async(ward=self.client.main_ward, variables= {"template": template_id, "provision": reference})
        logger.warn(f"Created {pod}")
        await self.client.entertainer.entertain(pod, actor)
        return pod


    async def provide_async(self):
        await self.setup_and_run()

    def provide(self):
        if self.loop.is_running():
            logger.error("Cannot do this, please await run()")
        else:
            self.loop.run_forever()

        # we enter a never-ending loop that waits for data
        # and runs callbacks whenever necessary.
        




from pydantic.main import BaseModel
from ....messages.types import  ASSIGN, PROVIDE
from ....messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from typing import List, Optional

class AssignParams(BaseModel):
    providers: Optional[List[str]]


class MetaExtensionsModel(MessageMetaExtensionsModel):
    with_progress: bool = False

class MetaModel(MessageMetaModel):
    type: str = ASSIGN
    extensions: Optional[MetaExtensionsModel]

class DataModel(MessageDataModel):
    node: Optional[str] #TODO: Maybe not optional
    template: Optional[str]
    pod: Optional[str]

    reservation: Optional[str]

    params: Optional[AssignParams]

    args: Optional[dict]
    kwargs: Optional[dict]


class AssignMessage(MessageModel):
    data: DataModel
    meta: MetaModel


from pydantic.main import BaseModel
from bergen.auths.legacy.app import LegacyApplication
from bergen.auths.backend.app import ArnheimBackendOauth
from bergen.auths.implicit.app import ImplicitApplication
from bergen.clients.base import ArkitektConfig, BaseBergen
from bergen.auths.types import GrantType, HerreConfig
import os
import yaml
import logging


logger = logging.getLogger(__name__)




class Bergen(BaseBergen):

    def __init__(self,
    config_path = "bergen.yaml",
    arkitekt_host: str = None, 
    arkitekt_port: int = None,
    herre_host: str = None,
    herre_port: int = None,
    client_id: str = None, 
    client_secret: str = None,
    username: str = None,
    password: str = None,
    grant_type: GrantType = GrantType.BACKEND,
    bind=True,
    allow_insecure=False,
    **kwargs) -> None:


        arkitekt_config = {}
        herre_config = {}

        if os.path.isfile(config_path):
            with open(config_path,"r") as file:
                logger.info("Using local configuration ")
                config = yaml.load(file, Loader=yaml.FullLoader)

                if "arkitekt" in config:
                    arkitekt_config.update(config["arkitekt"])

                if "herre" in config:
                    herre_config.update(config["herre"])


        if arkitekt_host : arkitekt_config["port"]= arkitekt_host
        if arkitekt_port : arkitekt_config["host"]= arkitekt_port
        if herre_host : herre_config["host"]= herre_host
        if herre_port : herre_config["port"]= herre_port
        if client_id : herre_config["client_id"]= client_id
        if client_secret : herre_config["client_secret"]= client_secret
        if grant_type: herre_config["grant_type"] = grant_type

        
        if allow_insecure: 
            herre_config["secure"] = False
            arkitekt_config["secure"] = False
                
        herre_config = HerreConfig(**herre_config)
        arkitekt_config = ArkitektConfig(**arkitekt_config)
        

        if grant_type == GrantType.BACKEND: auth = ArnheimBackendOauth(herre_config)
        elif grant_type == GrantType.IMPLICIT: auth = ImplicitApplication(herre_config)
        elif grant_type == GrantType.PASSWORD: auth = LegacyApplication(herre_config, username=username, password=password)
        else: raise NotImplementedError("Please Specifiy a valid Grant Type")

        super().__init__(auth, arkitekt_config, auto_negotiate=True, bind=bind, **kwargs)


from asyncio.tasks import ensure_future
from bergen.messages.postman.reserve.reserve_done import ReserveDoneMessage
from bergen.messages.postman.assign.assign_yield import AssignYieldsMessage
from bergen.messages.postman.assign.assign_critical import AssignCriticalMessage
from bergen.messages.postman.assign import AssignReturnMessage, AssignProgressMessage, CancelAssignMessage, AssignMessage
from bergen.messages.utils import expandToMessage
from bergen.messages.postman.provide import UnProvideMessage, ProvideDoneMessage, ProvideMessage,  ProvideProgressMessage, ProvideCriticalMessage
from bergen.messages.postman.reserve import ReserveMessage, CancelReserveMessage, UnReserveMessage, ReserveProgressMessage, ReserveCriticalMessage
from typing import Callable
from bergen.utils import expandOutputs, shrinkInputs
from bergen.messages.exception import ExceptionMessage
from bergen.messages.base import MessageModel
from bergen.postmans.base import BasePostman
import uuid
import logging
from functools import partial
import json

import asyncio
try:
    from asyncio import create_task
except ImportError:
    #python 3.6 fix
    create_task = asyncio.ensure_future


import websockets
from bergen.schema import AssignationStatus, Node, ProvisionStatus, Template
from bergen.models import Pod



logger = logging.getLogger(__name__)


class NodeException(Exception):
    pass

class PodException(Exception):
    pass

class ProviderException(Exception):
    pass


class Channel:

    def __init__(self, queue) -> None:
        self.queue = queue
        pass



class WebsocketPostman(BasePostman):
    type = "websocket"

    def __init__(self, port= None, protocol = None, host= None, auth= None, **kwargs) -> None:
        self.token = auth["token"]
        self.port = port
        self.protocol = protocol
        self.host = host
        self.connection = None      
        self.channel = None         
        self.callback_queue = ''

        self.uri = f"ws://{self.host}:{self.port}/postman/?token={self.token}"
        self.progresses = {}

        # Retry logic
        self.allowed_retries = 2
        self.current_retries = 0

        # Result and Stream Function
        self.futures = {}
        self.streams = {}   # Are queues that are consumed by tasks
        
        # Progress
        self.progresses = {}  # Also queues
        self.pending = []


        self.assign_routing = "assignation_request"
        super().__init__(**kwargs)

    async def connect(self):
        self.callback_queue = asyncio.Queue()
        self.progress_queue = asyncio.Queue()
        self.send_queue = asyncio.Queue()


        self.tasks = []

        self.startup_task = create_task(self.startup())


    async def disconnect(self):

        for task in self.pending:
            task.cancel()

        if self.connection: await self.connection.close()
        if self.receiving_task: self.receiving_task.cancel()
        if self.sending_task: self.sending_task.cancel()
        if self.callback_task: self.callback_task.cancel()

        if self.startup_task:
            self.startup_task.cancel()

        try:
            await self.startup_task
        except asyncio.CancelledError:
            logger.info("Postman disconnected")

    async def startup(self):
        try:
            await self.connect_websocket()
        except Exception as e:
            logger.debug(e)
            self.current_retries += 1
            if self.current_retries < self.allowed_retries:
                sleeping_time = (self.current_retries + 1)
                logger.error(f"Connection to Arkitekt Failed: Trying again in {sleeping_time} seconds.")
                await asyncio.sleep(sleeping_time)
                await self.startup()
            else:
                return

        self.receiving_task = create_task(
            self.receiving()
        )

        self.sending_task = create_task(
            self.sending()
        )

        self.callback_task = create_task(
            self.callbacks()
        )


        done, self.pending = await asyncio.wait(
            [self.callback_task, self.receiving_task, self.sending_task],
            return_when=asyncio.FIRST_EXCEPTION
        )

        logger.debug(f"Postman: Lost connection inbetween everything :( {[ task.exception() for task in done]}")
        logger.error(f'Postman: Trying to reconnect Postman')


        if self.connection: await self.connection.close()

        for task in self.pending:
            task.cancel()

        self.current_retries = 0 # reset retries after one successfull connection
        await self.startup()


    async def connect_websocket(self):
        self.connection = await websockets.client.connect(self.uri)
        logger.info("Successfully connected Postman")
        

    async def receiving(self):
        async for message in self.connection:
            await self.callback_queue.put(message)
    
    async def sending(self):
        while True:
            message = await self.send_queue.get()
            if self.connection:
                await self.connection.send(message.to_channels())
            else:
                raise Exception("No longer connected. Did you use an Async context manager?")

            self.send_queue.task_done()

    async def callbacks(self):
        while True:
            message = await self.callback_queue.get()
            logger.info(f"Received message {message}")
            try:
                parsed_message = expandToMessage(json.loads(message))
                correlation_id = parsed_message.meta.reference
                assert correlation_id in self.futures or correlation_id in self.streams , f"Received Message that wasn't originaing from our futures or streams {correlation_id}"
               
                # Stream Path

                if correlation_id in self.streams:
                    # It is a stream delage to stream
                    await self.streams[correlation_id].put(parsed_message)


                # Callback Path
                elif correlation_id in self.futures:
            
                    if isinstance(parsed_message, ExceptionMessage):
                        future = self.futures.pop(parsed_message.meta.reference)
                        future.set_exception(parsed_message.toException())

                    elif isinstance(parsed_message, AssignReturnMessage):
                        future = self.futures.pop(correlation_id)
                        future.set_result(parsed_message.data.returns)

                    elif isinstance(parsed_message, AssignProgressMessage):
                        if correlation_id in self.progresses:
                            self.progresses[correlation_id](parsed_message.data.message) 

                    elif isinstance(parsed_message, AssignCriticalMessage):
                        future = self.futures.pop(correlation_id)
                        future.set_exception(NodeException(parsed_message.data.message))

                    elif isinstance(parsed_message, ProvideDoneMessage):
                        future = self.futures.pop(correlation_id)
                        future.set_result(parsed_message.data.pod)

                    elif isinstance(parsed_message, ReserveDoneMessage):
                        future = self.futures.pop(correlation_id)
                        future.set_result((parsed_message.meta.reference, parsed_message.data.channel))

                    elif isinstance(parsed_message, ProvideCriticalMessage):
                        future = self.futures.pop(correlation_id)
                        future.set_exception(ProviderException(parsed_message.data.message))

                    elif isinstance(parsed_message, ProvideProgressMessage):
                        if correlation_id in self.progresses:
                            self.progresses[correlation_id](parsed_message.data.message) # call the function that is the progress function
 

                    else:
                        raise Exception("Unknown message type", parsed_message )


                else:
                    raise Exception("Message not in correlation_id", parsed_message )

            except Exception as e:
                logger.error(e)

            self.callback_queue.task_done()


    async def buildAssignMessage(self, reference: str = None, node: Node = None, pod: Pod = None, reservation: str = None, args=None, kwargs = None, params= None, with_progress=False):
        assert reference is not None, "Must have a reference"

        logger.info(args, kwargs)
        try:
            args, kwargs = await shrinkInputs(node=node, args=args, kwargs=kwargs)
        except Exception as e:
            logger.error(e)
        assign =  AssignMessage(data={
                                    "node": node.id if node else None, 
                                    "pod": pod.id if pod else None, 
                                    "reservation": reservation,
                                    "args": args, 
                                    "kwargs": kwargs,
                                    "params": dict(params or {}),
                                },
                                meta={
                                    "reference": reference,
                                    "extensions": {
                                        "progress": reference if with_progress else None,
                                        "callback": reference
                                    }
                                })
        logger.info(assign)
        print(assign)
        return assign


    async def stream(self, node: Node = None, pod: Pod = None, reservation: str = None, args = None, kwargs = None, params= None, on_progress: Callable = None):
        
        reference = str(uuid.uuid4())
        logger.info(f"Creating a Stream {reference}")
        self.streams[reference] = asyncio.Queue()

        with_progress = False
        if on_progress:
            assert callable(on_progress), "on_progress if provided must be a function/lambda"
            with_progress = True
        
        assign = await self.buildAssignMessage(reference=reference, node=node, pod=pod, reservation=reservation, args=args, kwargs=kwargs, params=params, with_progress=with_progress)
        await self.send_to_arkitekt(assign)

        while True:
            parsed_message = await self.streams[reference].get()

            if isinstance(parsed_message, ExceptionMessage):
                raise parsed_message.toException()

            if isinstance(parsed_message, AssignProgressMessage):
                if on_progress:
                    asyncio.get_event_loop().call_soon_threadsafe(on_progress(parsed_message.data.message))

            if isinstance(parsed_message, AssignYieldsMessage):
                yield await expandOutputs(node, outputs=parsed_message.data.yields)

            if isinstance(parsed_message, AssignReturnMessage):
                break


        

    async def send_to_arkitekt(self,request: MessageModel):
        await self.send_queue.put(request)

    async def assign(self, node: Node = None, pod: Pod = None, reservation: str = None, args = None, kwargs = None, params= None, on_progress: Callable = None):
        reference = str(uuid.uuid4()) 
        # Where should we do this?
        future = self.loop.create_future()
        self.futures[reference] = future

        with_progress = False
        if on_progress:
            assert callable(on_progress), "on_progress if provided must be a function/lambda"
            self.progresses[reference] = lambda progress: logger.info(progress) 
            with_progress = True
        
        assign = await self.buildAssignMessage(reference=reference, node=node, pod=pod, reservation=reservation, args=args, kwargs=kwargs, params=params, with_progress=with_progress)
        await self.send_to_arkitekt(assign)
        logger.warn(assign)

        future.add_done_callback(partial(self.check_if_assign_cancel, reference))  
        outputs = await future

        return await expandOutputs(node, outputs)


    def check_if_assign_cancel(self, reference, future):
        # We cancelled the future and now would like to cancel it also on the Arnheim side
        if future.cancelled():
            cancel = CancelAssignMessage(data={reference: reference}, meta={reference: reference})
            self.tasks.append(asyncio.create_task(self.send_to_arkitekt(cancel)))
            logger.warn(f"Cancelling Assignation {cancel.data.reference}")

        if future.exception():
            raise future.exception()

    def check_if_provide_cancel(self, reference, future):
        # We cancelled the future and now would like to cancel it also on the Arnheim side
        if future.cancelled():
            cancel = CancelAssignMessage(data={reference: reference}, meta={reference: reference})
            self.tasks.append(asyncio.create_task(self.send_to_arkitekt(cancel)))
            logger.warn(f"Cancelling Assignation {cancel.data.reference}")

        if future.exception():
            raise future.exception()


    def check_if_reserve_cancel(self, reference, future):
        # We cancelled the future and now would like to cancel it also on the Arnheim side
        if future.cancelled():
            cancel = CancelReserveMessage(data={reference: reference}, meta={reference: reference})
            self.tasks.append(asyncio.create_task(self.send_to_arkitekt(cancel)))
            logger.warn(f"Cancelling Reservation {cancel.data.reference}")

        if future.exception():
            raise future.exception()


    async def buildProvideMessage(self, reference: str = None, node: Node = None, template: Template = None, params= None, with_progress=False):
        assert reference is not None, "Must have a reference"

        return ProvideMessage(data={
                                    "node": node.id if node else None, 
                                    "template": template.id if template else None, 
                                    "params": dict(params or {}),
                                },
                                meta={
                                    "reference": reference,
                                    "extensions": {
                                        "progress": reference if with_progress else None,
                                        "callback": reference
                                    }
      
                                })

    async def buildReserveMessage(self, reference: str = None, node: Node = None, template: Template = None, params= None, with_progress=False):
        assert reference is not None, "Must have a reference"

        return ReserveMessage(data={
                                    "node": node.id if node else None, 
                                    "template": template.id if template else None, 
                                    "params": dict(params or {}),
                                },
                                meta={
                                    "reference": reference,
                                    "extensions": {
                                        "progress": reference if with_progress else None,
                                        "callback": reference
                                    }
      
                                })



    async def buildUnProvideMessage(self, reference: str = None, pod: Pod = None,  with_progress=False):
        assert pod is not None, "Must have a pod to unprovide"

        return UnProvideMessage(data={
                                    "pod": pod.id , 
                                },
                                meta={
                                    "reference": reference,
                                    "extensions": {
                                        "progress": reference if with_progress else None,
                                        "callback": reference
                                    }

                                })


    async def buildUnReserveMessage(self, reference: str = None, reservation: str = None,  with_progress=False):
        assert reservation is not None, "Must have a reservation to unreserve"

        return UnReserveMessage(data={
                                    "reservation": reservation , 
                                },
                                meta={
                                    "reference": reference,
                                    "extensions": {
                                        "progress": reference if with_progress else None,
                                        "callback": reference
                                    }

                                })


    async def unprovide(self, pod: Pod= None, on_progress: Callable = None) -> Pod:
        reference = str(uuid.uuid4()) 
        # Where should we do this?
        future = self.loop.create_future()
        self.futures[reference] = future

        with_progress = False
        if on_progress:
            assert callable(on_progress), "on_progress if provided must be a function/lambda"
            self.progresses[reference] = lambda progress: logger.info(progress) 
            with_progress = True
        
        assign = await self.buildUnProvideMessage(reference=reference, pod=pod, with_progress=with_progress)
        await self.send_to_arkitekt(assign)
        result = await future
        return result


    async def provide(self, node: Node = None, template: Template = None , params= None, on_progress: Callable = None) -> Pod:
        reference = str(uuid.uuid4()) 
        # Where should we do this?
        future = self.loop.create_future()
        self.futures[reference] = future

        with_progress = False
        if on_progress:
            assert callable(on_progress), "on_progress if provided must be a function/lambda"
            self.progresses[reference] = lambda progress: logger.info(progress) 
            with_progress = True
        
        assign = await self.buildProvideMessage(reference=reference, node=node, template=template, params=params, with_progress=with_progress)
        await self.send_to_arkitekt(assign)

        future.add_done_callback(partial(self.check_if_provide_cancel, reference))  
        id = await future

        pod = await Pod.asyncs.get(id=id)
        return pod


    async def reserve(self, node: Node = None, template: Template = None , params= None, on_progress: Callable = None) -> str:
        reference = str(uuid.uuid4()) 
        # Where should we do this?
        future = self.loop.create_future()
        self.futures[reference] = future

        with_progress = False
        if on_progress:
            assert callable(on_progress), "on_progress if provided must be a function/lambda"
            self.progresses[reference] = lambda progress: logger.info(progress) 
            with_progress = True
        
        assign = await self.buildReserveMessage(reference=reference, node=node, template=template, params=params, with_progress=with_progress)
        await self.send_to_arkitekt(assign)

        future.add_done_callback(partial(self.check_if_reserve_cancel, reference))  
        reservation_id = await future
        return reservation_id

    async def unreserve(self, reservation: str= None, on_progress: Callable = None) -> Pod:
        reference = str(uuid.uuid4()) 
        # Where should we do this?
        future = self.loop.create_future()
        self.futures[reference] = future

        with_progress = False
        if on_progress:
            assert callable(on_progress), "on_progress if provided must be a function/lambda"
            self.progresses[reference] = lambda progress: logger.info(progress) 
            with_progress = True
        
        assign = await self.buildUnReserveMessage(reference=reference, reservation=reservation, with_progress=with_progress)
        await self.send_to_arkitekt(assign)
        result = await future
        return result


    async def reserve_stream(self, node: Node = None, template: Template = None , params= None, on_progress: Callable = None):
        reference = str(uuid.uuid4())
        logger.info(f"Creating a Reservation Stream {reference}")

        self.streams[reference] = asyncio.Queue()

        with_progress = False
        if on_progress:
            assert callable(on_progress), "on_progress if provided must be a function/lambda"
            with_progress = True
        
        assign = await self.buildReserveMessage(reference=reference, node=node, template=template, params=params, with_progress=with_progress)
        await self.send_to_arkitekt(assign)

        while True:
            parsed_message = await self.streams[reference].get()
            yield parsed_message



from abc import ABC, abstractmethod

from requests.models import Response
from bergen.enums import ClientType
import shelve
import os
import requests
import logging
from .types import User, HerreConfig

logger = logging.getLogger(__name__)



class AuthError(Exception):
    pass





class BaseAuthBackend(ABC):


    def __init__(self, config: HerreConfig, token_url="o/token/", authorize_url="o/authorize/", check_endpoint="auth/") -> None:
        # Needs to be set for oauthlib
        os.environ["OAUTHLIB_INSECURE_TRANSPORT"] = "0" if config.secure else "1"



        self.base_url = f'{"https" if config.secure else "http"}://{config.host}:{config.port}/'
        self.check_url = self.base_url + check_endpoint
        self.auth_url = self.base_url + authorize_url
        self.token_url = self.base_url + token_url
        self.scopes = config.scopes + ["introspection"]
        self.client_id = config.client_id
        self.client_secret = config.client_secret  

        self.scope = " ".join(self.scopes)
        self._user = None

        # Lets check if we already have a local toke
        config_name = "token.db"
        run_path = os.path.abspath(os.getcwd())
        self.config_path = os.path.join(run_path, config_name)

        try:
            with shelve.open(self.config_path) as cfg:
                    self.token = cfg['token']
                    self.needs_validation = True
                    logger.debug("Found local config")
        except KeyError:
            self.token = None
            self.needs_validation = False

        super().__init__()


    @abstractmethod
    def fetchToken(self, loop=None) -> str:
        raise NotImplementedError("This is an abstract Class")

    def refetchToken(self) -> str:
        raise NotImplementedError("This is an abstract Class")

    def getUser(self):
        assert self.token is not None, "Need to authenticate before accessing the User"
        if not self._user:
            answer = requests.get(self.base_url + "me", headers={"Authorization": f"Bearer {self.token}"})
            self._user = User(**answer.json())
        return self._user

    def getToken(self, loop=None) -> str:
        if self.token is None:
            self.token = self.fetchToken()
            
            with shelve.open(self.config_path) as cfg:
                cfg['token'] = self.token

            return self.token
        
        else:
            if self.needs_validation:
                response = requests.post(self.check_url, {"token": self.token}, headers={"Authorization": f"Bearer {self.token}"})
                print(response.status_code)
                self.needs_validation = False
                if response.status_code == 200:
                    logger.info("Old token still valid!")
                    return self.token
                else:
                    logger.info("Need to refetch Token!!") # Was no longer valid, fetching anew
                    self.token = self.fetchToken()

                    with shelve.open(self.config_path) as cfg:
                        cfg['token'] = self.token

                    return self.token
                
            return self.token




import asyncio
from asyncio.futures import Future
from bergen.messages.exception import ExceptionMessage
from bergen.messages.postman.reserve.reserve_critical import ReserveCriticalMessage
from typing import Any
from bergen.schema import AssignationParams, Pod, PodStatus, ProvisionParams
from bergen.registries.arnheim import get_current_arnheim
from bergen.types.model import ArnheimModel
from bergen.extenders.base import BaseExtender
from aiostream import stream
from bergen.extenders.contexts.pod import HostedPod
from bergen.messages.postman.reserve import ReserveDoneMessage
from tqdm import tqdm
import textwrap
import logging

logger = logging.getLogger(__name__)

class AssignationUIMixin:

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._ui = None


    def askInputs(self, **kwargs) -> dict:
        widget = self.getWidget(**kwargs) # We have established a ui
        if widget.exec_():
            return widget.parameters
        else:
            return None


    def getWidget(self, **kwargs):
        try:
            from bergen.ui.assignation import AssignationUI
            if not self._ui:
                self._ui = AssignationUI(self.inputs, **kwargs)
            return self._ui
        except ImportError as e:
            raise NotImplementedError("Please install PyQt5 in order to use interactive Widget based parameter query")
            
        


class ProvideContext:


    def __init__(self, node, on_progress=None, **params) -> None:
        bergen = get_current_arnheim()

        self._postman = bergen.getPostman()
        self.node = node
        self.on_progress = on_progress
        self.params = ProvisionParams(**params)
        pass


    async def assign(self, *args, **kwargs):
        return await self._postman.assign(pod=self.pod, node=self.node, args=args, kwargs=kwargs, on_progress=self.on_progress)


    async def unprovide(self):
        return await self._postman.unprovide(pod=self.pod, on_progress=self.on_progress)

    async def provide(self):
        return await self._postman.provide(node=self.node, params=self.params, on_progress=self.on_progress)

    async def __aenter__(self):
        logger.info(f"Providing this node {self.node} with {self.params}")
        self.pod = await self.provide()
        logger.warn(f"Provided Listener on {self.pod.channel}")
        return self

    async def __aexit__(self, *args, **kwargs):
        await self.unprovide()


class ReserveContext:


    def __init__(self, node, on_progress=None, **params) -> None:
        bergen = get_current_arnheim()
        self._postman = bergen.getPostman()

        self.node = node
        self.on_progress = on_progress
        self.params = ProvisionParams(**params)
        pass


    async def assign(self, *args, **kwargs):
        return await self._postman.assign(reservation=self.reservation, node=self.node, args=args, kwargs=kwargs, on_progress=self.on_progress)

    async def unreserve(self):
        return await self._postman.unreserve(reservation=self.reservation, on_progress=self.on_progress)

    async def reserve(self):
        return await self._postman.reserve(node=self.node, params=self.params, on_progress=self.on_progress)

    async def __aenter__(self):
        logger.info(f"Reserving this node {self.node} with {self.params}")
        self.reservation, self.pod_channel = await self.reserve()
        logger.warn(f"Recevied reservation {self.reservation} on channel {self.pod_channel}")
        return self

    async def __aexit__(self, *args, **kwargs):
        await self.unreserve()


class Reservation:


    def __init__(self, node, on_progress=None, loop=None, **params) -> None:
        bergen = get_current_arnheim()
        self._postman = bergen.getPostman()

        self.node = node
        self.on_progress = on_progress
        self.params = ProvisionParams(**params)

        self.loop = loop or asyncio.get_event_loop()
        # Status
        self.running = False
        self.critical_error = None
        self.recovering = False #TODO: Implement

        pass


    async def assign(self, *args, **kwargs):
        assert self.critical_error is None, f"Contract was broken because pod is critically errored. All requests will be cancelled {self.critical_error}"
        return await self._postman.assign(reservation=self.reservation, node=self.node, args=args, kwargs=kwargs, on_progress=self.on_progress)

    def stream(self, *args, **kwargs):
        assert self.critical_error is None, f"Contract was broken because pod is critically errored. All requests will be cancelled {self.critical_error}"
        return self._postman.stream(reservation=self.reservation, node=self.node, args=args, kwargs=kwargs, on_progress=self.on_progress)

    async def unreserve(self):
        return await self._postman.unreserve(reservation=self.reservation, on_progress=self.on_progress)

    async def contract_worker(self):
        self.running = True
        async for message in self._postman.reserve_stream(node=self.node, params=self.params, on_progress=self.on_progress):

            # Before here because Reserve Critical is actually an ExceptionMessage
            #TODO: Undo this
            if isinstance(message, ReserveCriticalMessage):
                logger.error(f"Reservation was stopped with a Critical Error {message.data.message}")
                self.critical_error = message

            elif isinstance(message, ExceptionMessage):
                self.contract_started.set_exception(message.toException())
                return

            elif isinstance(message, ReserveDoneMessage):
                self.contract_started.set_result((message.meta.reference, message.data.channel))

            


    def cancel_reservation(self, future: Future):
        if future.cancelled():
            logger.info("We have finished with our reservation now. Bye bye Pod")
            return
        elif future.exception():
            logger.info("We terminated with an Error")
            raise future.exception()
        elif future.done():
            logger.error("We finished which is weird!")
            return


    async def __aenter__(self):
        logger.info(f"Reserving this node {self.node} with {self.params}")

        self.contract_started = self.loop.create_future()
        self.worker_task = self.loop.create_task(self.contract_worker())
        self.worker_task.add_done_callback(self.cancel_reservation)
        self.reservation, self.channel = await self.contract_started
        logger.warn(f"Recevied reservation {self.reservation} on channel {self.channel}")
        return self

    async def __aexit__(self, *args, **kwargs):
        if not self.worker_task.done():
            #await self._postman.unreserve(reservation=self.reservation, on_progress=self.on_progress)
            self.worker_task.cancel()
            try:
                await self.worker_task
            except asyncio.CancelledError:
                logger.info("Reservation over")

        

class NodeExtender(AssignationUIMixin, BaseExtender):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args,**kwargs)
        
        bergen = get_current_arnheim()

        self._postman = bergen.getPostman()
        self._loop, self._force_sync = bergen.getLoopAndContext()


    def provide(self, **params) -> ProvideContext:
        return ProvideContext(self, **params)

    def reserve(self, **params) -> Reservation:
        return Reservation(self, **params)

    async def provide_async(self, params: ProvisionParams, **kwargs):
        return await self._postman.provide(self, params, **kwargs)       

    async def assign_async(self, inputs: dict, params: AssignationParams, **kwargs):
        
        return await self._postman.assign(self, inputs, params, **kwargs)

    async def delay_async(self, inputs: dict, params: AssignationParams, **kwargs):
    
        return await self._postman.delay(self, inputs, params, **kwargs)

    def stream(self, inputs: dict, params: AssignationParams = None, **kwargs):

        return stream.iterate(self._postman.stream(self, inputs, params, **kwargs))


    async def assign_with_progress(self, inputs, params, **kwargs):
        result = None
        with tqdm(total=100) as pbar:
                async with self.stream_progress(inputs, params, **kwargs) as stream:
                        async for item in stream:
                                result = item
                                if isinstance(result, dict): break
                                
                                progress, message = item.split(":")
                                try: 
                                        pbar.n = int(progress)
                                        pbar.refresh()
                                except:
                                        pass
                                pbar.set_postfix_str(textwrap.shorten(message, width=30, placeholder="..."))
                pbar.n = 100
                pbar.refresh()
                pbar.set_postfix_str("Done")
        return result


    def stream_progress(self,  inputs: dict, params: AssignationParams = None, **kwargs):
        return stream.iterate(self._postman.stream_progress(self, inputs, params, **kwargs)).stream()
    
    def delay(self, inputs: dict, params: AssignationParams = None, **kwargs):
        if self._loop.is_running() and not self._force_sync:
            return self.delay_async(inputs, params, **kwargs)
        else:
            result = self._loop.run_until_complete(self.delay_async(inputs, params, **kwargs))
            return result


    def __call__(self, inputs: dict, params: AssignationParams = None, with_progress = False, **kwargs) -> dict:
        """Call this node (can be run both asynchronously and syncrhounsly)

        Args:
            inputs (dict): The inputs for this Node
            params (AssignationParams, optional): [description]. Defaults to None.

        Returns:
            outputs (dict): The ooutputs of this Node
        """
    
        if self._loop.is_running() and not self._force_sync:
            if with_progress == True:
                return self.assign_with_progress(inputs, params,  with_progress=with_progress, **kwargs)
            return self.assign_async(inputs, params, with_progress=with_progress, **kwargs)

        else:
            if with_progress == True:
                return self._loop.run_until_complete(self.assign_with_progress(inputs, params,  with_progress=with_progress, **kwargs))

            result = self._loop.run_until_complete(self.assign_async(inputs, params,  with_progress=with_progress, **kwargs))
            return result



    def hosted(self, params: ProvisionParams = {}, enter_when = PodStatus.ACTIVE, **kwargs):
        return HostedPod(self, params = {}, enter_when=enter_when, **kwargs)

    def _repr_html_(self):
        string = f"{self.name}</br>"

        for input in self.inputs:
            string += "Inputs </br>"
            string += f"Port: {input._repr_html_()} </br>"

        for output in self.outputs:
            string += "Outputs </br>"
            string += f"Port: {output._repr_html_()} </br>"


        return string


from pydantic.main import BaseModel
from ....messages.types import  UNRESERVE
from ....messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from typing import List, Optional



class MessageMetaExtensionsModel(MessageMetaExtensionsModel):
    # Set by postman consumer
    progress: Optional[str]
    callback: Optional[str]

class MetaModel(MessageMetaModel):
    type: str = UNRESERVE
    extensions: Optional[MessageMetaExtensionsModel]

class DataModel(MessageDataModel):
    reservation: Optional[str] 

class UnReserveMessage(MessageModel):
    data: DataModel
    meta: MetaModel


from ....messages.types import CANCEL_RESERVE
from ....messages.base import MessageDataModel, MessageMetaExtensionsModel, MessageMetaModel, MessageModel
from typing import Optional


class CancelProvideMetaModel(MessageMetaModel):
    type: str = CANCEL_RESERVE

class CancelProvideDataModel(MessageDataModel):
    reference: str


class CancelReserveMessage(MessageModel):
    data: CancelProvideDataModel
    meta: CancelProvideMetaModel


from .bounced_cancel_reserve import BouncedCancelReserveMessage
from .bounced_reserve import BouncedReserveMessage
from .cancel_reserve import CancelReserveMessage
from .reserve_critical import ReserveCriticalMessage
from .reserve_done import ReserveDoneMessage
from .reserve_progress import ReserveProgressMessage
from .reserve import ReserveMessage
from .unreserve import UnReserveMessage


import asyncio
from abc import ABC, abstractmethod
from asyncio.futures import Future
import logging
from typing import Dict, Type
from bergen.models import Pod
from .actor import Actor, HostHelper
from functools import partial
import uuid

logger = logging.getLogger(__name__)

class BaseHost(ABC):
    ''' Is a mixin for Our Bergen '''
    helperClass = None

    def __init__(self, raise_exceptions_local=False, client = None, loop=None,**kwargs) -> None:
        self.pods = {}
        self.raise_exceptions_local = raise_exceptions_local
        self.loop = loop or asyncio.get_event_loop()
        self.client = client

        self.hostHelper = self.helperClass(self)

        self.pod_actor_run_map: Dict[str, asyncio.Task]= {}
        self.pod_actor_queue_map: Dict[str, asyncio.Queue] = {}

        self.entertainments = {}
        self.assignments = {}

        self.pending = []

        self.tasks = []

    @abstractmethod
    async def connect(self) -> str:
        raise NotImplementedError("Please overwrite")

    @abstractmethod
    async def disconnect(self) -> str:
        raise NotImplementedError("Please overwrite")


    @abstractmethod
    async def activatePod(self):
         raise NotImplementedError("Please overwrite")

    @abstractmethod
    async def deactivatePod(self):
         raise NotImplementedError("Please overwrite")

    async def unentertain(self, pod_id: str):
        assert pod_id not in self.pod_actor_run_map, "This actor is not entertained"
        task = self.pod_actor_run_map.pop(pod_id)
        if task.done():
            raise Exception("Task was already done, this is weird")
        elif task.exception():
            logger.info("Task already was cancelled apparently")
        else:
            task.cancel()
    

    def actor_cancelled(self, actor: Actor, future: Future):
        logger.info("Actor is done! Cancellation or Finished??")
        if future.exception():
            raise future.exception()


        self.tasks.append(asyncio.create_task(self.deactivatePod(actor.pod)))

    async def entertain(self, pod: Pod, actorClass: Type[Actor]):
        ''' Takes an instance of a pod, asks arnheim to activate it and accepts requests on it,
        cancel this task to unprovide your local implementatoin '''
        assert pod.id not in self.pod_actor_run_map, "This pod is already entertained"

        reference = str(uuid.uuid4())
        actor = actorClass(pod, self.hostHelper)
        await self.activatePod(pod, reference)

        try:
            self.pod_actor_queue_map[str(pod.id)] = actor.queue
            run_task = asyncio.create_task(actor.run())
            run_task.add_done_callback(partial(self.actor_cancelled, actor))
            self.pod_actor_run_map[str(pod.id)] = run_task

        except Exception as e:
            logger.error(e)




from abc import ABC, abstractmethod
from bergen.schema import Node, Pod
from typing import Callable
from aiostream import stream
import asyncio

class BasePostman(ABC):
    """ A Postman takes node requests and translates them to Bergen calls, basic implementations are GRAPHQL and PIKA"""

    def __init__(self, requires_configuration=True, loop=None, client=None) -> None:
        assert loop is not None, "Please provide a Loop to your Postman, Did you forget to call super.init with **kwargs?"
        self.loop = loop or asyncio.get_event_loop()
        self.client = client

    @abstractmethod
    async def connect(self):
        pass

    @abstractmethod
    async def disconnect(self):
        pass

    @abstractmethod
    async def stream(self, node: Node = None, pod: Pod = None, args = None, kwargs = None, params= None, on_progress: Callable = None):
        return NotImplementedError( "Abstract class")

    @abstractmethod
    async def assign(self, node: Node = None, pod: Pod = None, reservation: str = None,  args = None, kwargs = None, params= None, on_progress: Callable = None):
        return NotImplementedError("This is abstract")

    @abstractmethod
    async def provide(self, *args, **kwargs):
        return NotImplementedError("This is abstract")

    @abstractmethod
    async def unprovide(self, *args, **kwargs):
        return NotImplementedError("This is abstract")




